part of maps;

Future<Map<String, dynamic>> _retrieveDataFromShapeFile(String file,
    Map<String, dynamic> cacheData, bool isShapeFileDecoded) async {
  if (isShapeFileDecoded) {
    return cacheData;
  }
  final String data = await rootBundle.loadString(file);
  return compute(_decodeJsonData, data);
}

Map<String, dynamic> _decodeJsonData(String data) {
  return jsonDecode(data);
}

Size _getActualSize(BoxConstraints constraints) {
  final double width = constraints.hasBoundedWidth ? constraints.maxWidth : 300;
  final double height =
      constraints.hasBoundedHeight ? constraints.maxHeight : 300;
  return Size(width, height);
}

/// Convert latitude and longitude co-ordinate points to pixel points.
Offset _getPixelPositionFromCoordinate(num longitude, num latitude, Size size,
    {Offset translationOffset = const Offset(0, 0), double factor = 1.0}) {
  assert(longitude != null);
  assert(latitude != null);
  assert(size != null);

  final num x = (longitude + 180) / 360;
  final num sinLatitude = sin(latitude * pi / 180);
  final num y = 0.5 - log((1 + sinLatitude) / (1 - sinLatitude)) / (4 * pi);
  final num mapSize = size.longestSide * factor;
  final num dx = translationOffset.dx + _clip(x * mapSize, 0, mapSize);
  final num dy = translationOffset.dy + _clip(y * mapSize, 0, mapSize);
  return Offset(dx, dy);
}

/// Get boundary value.
double _clip(double value, double minValue, double maxValue) {
  assert(value != null);
  assert(minValue != null);
  assert(maxValue != null);
  return min(max(value, minValue), maxValue);
}
