part of maps;

/// A Render object widget which draws bubble shape.
class _MapBubble extends LeafRenderObjectWidget {
  const _MapBubble({this.bubbleSettings, this.state});

  final MapBubbleSettings bubbleSettings;

  final _MapsShapeLayerState state;

  @override
  _RenderMapBubble createRenderObject(BuildContext context) {
    return _RenderMapBubble(
        bubbleSettings: bubbleSettings,
        mediaQueryData: MediaQuery.of(context),
        state: state);
  }

  @override
  void updateRenderObject(BuildContext context, _RenderMapBubble renderObject) {
    renderObject
      ..bubbleSettings = bubbleSettings
      ..mediaQueryData = MediaQuery.of(context);
  }
}

class _RenderMapBubble extends RenderBox {
  _RenderMapBubble({
    MapBubbleSettings bubbleSettings,
    MediaQueryData mediaQueryData,
    _MapsShapeLayerState state,
  })  : _bubbleSettings = bubbleSettings,
        _mediaQueryData = mediaQueryData,
        _state = state {
    _bubbleAnimation = CurvedAnimation(
        parent: _state.bubbleAnimationController,
        curve: const Interval(0.2, 1.0, curve: Curves.easeInOut));
  }

  Animation<double> _bubbleAnimation;

  final _MapsShapeLayerState _state;

  MapBubbleSettings get bubbleSettings => _bubbleSettings;
  MapBubbleSettings _bubbleSettings;

  set bubbleSettings(MapBubbleSettings value) {
    if (_bubbleSettings == value) {
      return;
    }
    _bubbleSettings = value;
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;

  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    markNeedsLayout();
  }

  @override
  bool get isRepaintBoundary => true;

  @override
  void performLayout() {
    size = _getActualSize(constraints);
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _bubbleAnimation.addListener(markNeedsPaint);
  }

  @override
  void detach() {
    _bubbleAnimation.removeListener(markNeedsPaint);
    super.detach();
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    _drawBubbleShape(context, offset);
  }

  void _drawBubbleShape(PaintingContext context, Offset offset) {
    final _MapChildParentData childParentData = parentData;
    if (childParentData.mapDataSource != null) {
      final Paint paint = Paint()..isAntiAlias = true;
      childParentData.mapDataSource.forEach((String key, _MapModel model) {
        if (model.bubbleSizeValue == null) {
          return;
        }

        final double bubbleRadius = _bubbleAnimation.value * model.bubbleRadius;
        final double opacity = _bubbleAnimation.value * _bubbleSettings.opacity;
        // Drawing bubble shape
        paint
          ..style = PaintingStyle.fill
          ..color =
              model.bubbleColor ?? bubbleSettings.color.withOpacity(opacity);
        if (paint.color != null && paint.color != Colors.transparent) {
          context.canvas.drawCircle(model.shapePathCenter, bubbleRadius, paint);
        }

        if (_bubbleSettings.strokeColor != null &&
            _bubbleSettings.strokeWidth != null) {
          final double strokeWidth = _bubbleSettings.strokeWidth;
          paint
            ..style = PaintingStyle.stroke
            ..color = _bubbleSettings.strokeColor
            ..strokeWidth =
                strokeWidth > bubbleRadius ? bubbleRadius : strokeWidth;
          // Drawing bubble stroke
          context.canvas.drawCircle(
              model.shapePathCenter,
              strokeWidth > bubbleRadius
                  ? bubbleRadius / 2
                  : bubbleRadius - strokeWidth / 2,
              paint);
        }
      });
    }
  }
}
