part of maps;

/// A Render object widget which renders all data labels.
class _MapDataLabel extends LeafRenderObjectWidget {
  const _MapDataLabel(
      {this.dataLabelSettings, this.defautTextStyle, this.state});

  final MapDataLabelSettings dataLabelSettings;

  final TextStyle defautTextStyle;

  final _MapsShapeLayerState state;

  @override
  _RenderMapDataLabel createRenderObject(BuildContext context) {
    return _RenderMapDataLabel(
      dataLabelSettings: dataLabelSettings,
      defautTextStyle: defautTextStyle,
      mediaQueryData: MediaQuery.of(context),
      state: state,
    );
  }

  @override
  void updateRenderObject(
      BuildContext context, _RenderMapDataLabel renderObject) {
    renderObject
      ..dataLabelSettings = dataLabelSettings
      ..mediaQueryData = MediaQuery.of(context);
  }
}

class _RenderMapDataLabel extends RenderBox {
  _RenderMapDataLabel({
    MapDataLabelSettings dataLabelSettings,
    TextStyle defautTextStyle,
    MediaQueryData mediaQueryData,
    _MapsShapeLayerState state,
  })  : _dataLabelSettings = dataLabelSettings,
        _defaultTextStyle = defautTextStyle,
        _mediaQueryData = mediaQueryData,
        _state = state {
    _textPainter = TextPainter(textDirection: TextDirection.ltr);
    _textPainter.textScaleFactor = _mediaQueryData.textScaleFactor;
    _dataLabelAnimation = Tween<double>(begin: 0.0, end: 1.0).animate(
      CurvedAnimation(
        parent: _state.dataLabelAnimationController,
        curve: const Interval(0.2, 1.0, curve: Curves.easeInOut),
      )..addStatusListener(
          (AnimationStatus status) {
            if (status == AnimationStatus.completed &&
                _state.widget.showBubbles) {
              _state.bubbleAnimationController.forward();
            }
          },
        ),
    );
  }

  final _MapsShapeLayerState _state;

  final TextStyle _defaultTextStyle;

  TextPainter _textPainter;

  Animation<double> _dataLabelAnimation;

  MapDataLabelSettings get dataLabelSettings => _dataLabelSettings;
  MapDataLabelSettings _dataLabelSettings;

  set dataLabelSettings(MapDataLabelSettings value) {
    if (_dataLabelSettings == value) {
      return;
    }
    _dataLabelSettings = value;
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;

  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    _textPainter.textScaleFactor = _mediaQueryData.textScaleFactor;
    markNeedsPaint();
  }

  @override
  bool get isRepaintBoundary => true;

  @override
  void performLayout() {
    size = _getActualSize(constraints);
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _dataLabelAnimation.addListener(markNeedsPaint);
  }

  @override
  void detach() {
    _dataLabelAnimation.removeListener(markNeedsPaint);
    super.detach();
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    _drawDataLabel(context, offset);
  }

  void _drawDataLabel(PaintingContext context, Offset offset) {
    final _MapChildParentData childParentData = parentData;
    if (childParentData.mapDataSource != null) {
      String dataLabelText;
      bool hasTextStyle = false;
      TextStyle textStyle = _dataLabelSettings.textStyle ??
          _state.mapsThemeData.dataLabelTextStyle;
      if (textStyle != null) {
        hasTextStyle = true;
        textStyle = textStyle.copyWith(
            color: textStyle.color.withOpacity(_dataLabelAnimation.value));
      }

      final bool hasDataLabelMapper =
          _state.widget.delegate.dataLabelMapper != null;

      childParentData.mapDataSource.forEach((String key, _MapModel model) {
        dataLabelText =
            hasDataLabelMapper ? model.dataLabelText : model.primaryKey;
        if (dataLabelText == null) {
          return;
        }

        if (!hasTextStyle) {
          textStyle = _defaultTextStyle.copyWith(
              color: ((model.shapeColor ?? _state.mapsThemeData.layerColor)
                              .computeLuminance() >
                          0.5
                      ? Colors.black
                      : Colors.white)
                  .withOpacity(_dataLabelAnimation.value));
        }

        _textPainter.text = TextSpan(style: textStyle, text: dataLabelText);
        _textPainter.layout();

        if (_dataLabelSettings.overflowMode == MapLabelOverflowMode.hide) {
          if (_textPainter.width > model.shapeWidth) {
            return;
          }
        } else if (_dataLabelSettings.overflowMode ==
            MapLabelOverflowMode.trim) {
          final int actualTextLength = dataLabelText.length;
          String trimedText = dataLabelText;
          int trimLength = 3; // 3 dots
          while (_textPainter.width > model.shapeWidth) {
            if (trimedText.length <= 4) {
              _textPainter.text =
                  TextSpan(style: textStyle, text: trimedText[0] + '...');
              _textPainter.layout();
              break;
            } else {
              trimedText = dataLabelText.replaceRange(
                  actualTextLength - trimLength, actualTextLength, '...');
              _textPainter.text = TextSpan(style: textStyle, text: trimedText);
              _textPainter.layout();
              trimLength++;
            }
          }
        }

        _textPainter.paint(
            context.canvas,
            model.shapePathCenter -
                Offset(_textPainter.width / 2, _textPainter.height / 2));
      });
    }
  }
}
