part of maps;

class _RenderShapeLayer extends RenderStack {
  _RenderShapeLayer({
    Map<String, _MapModel> mapDataSource,
    MapShapeLayerDelegate mapDelegate,
    Color color,
    Color strokeColor,
    double strokeWidth,
    MapBubbleSettings bubbleSettings,
    MapLegendSettings legendSettings,
    MapSelectionSettings selectionSettings,
    MediaQueryData mediaQueryData,
    bool enableShapeTooltip,
    bool enableBubbleTooltip,
    bool enableSelection,
    SfMapsThemeData mapsThemeData,
    _MapsShapeLayerState state,
  })  : _mapDataSource = mapDataSource,
        _mapDelegate = mapDelegate,
        _color = color,
        _strokeColor = strokeColor,
        _strokeWidth = strokeWidth,
        _bubbleSettings = bubbleSettings,
        _legendSettings = legendSettings,
        _selectionSettings = selectionSettings,
        _mediaQueryData = mediaQueryData,
        _enableShapeTooltip = enableShapeTooltip,
        _enableBubbleTooltip = enableBubbleTooltip,
        _enableSelection = enableSelection,
        _mapsThemeData = mapsThemeData,
        _state = state,
        super(textDirection: Directionality.of(state.context)) {
    _tapGestureRecognizer = TapGestureRecognizer()..onTapUp = _onTapUp;
  }

  final _MapsShapeLayerState _state;

  double _shapeSizeFactor = 1.0;

  double _actualShapeHeight;

  double _actualShapeWidth;

  Offset _shapeLayerTranslationOffset = const Offset(0.0, 0.0);

  _MapModel _prevSelectedItem;

  _MapModel _currentSelectedItem;

  TapGestureRecognizer _tapGestureRecognizer;

  Size _currentSize;

  final bool _mouseIsConnected =
      RendererBinding.instance.mouseTracker.mouseIsConnected;

  bool get isInteractive =>
      _enableSelection || _enableBubbleTooltip || _enableShapeTooltip;

  Map<String, _MapModel> get mapDataSource => _mapDataSource;
  Map<String, _MapModel> _mapDataSource;
  set mapDataSource(Map<String, _MapModel> value) {
    if (const MapEquality<String, _MapModel>().equals(_mapDataSource, value)) {
      return;
    }

    _mapDataSource = value;
    _refresh();
    markNeedsPaint();
  }

  MapShapeLayerDelegate get mapDelegate => _mapDelegate;
  MapShapeLayerDelegate _mapDelegate;
  set mapDelegate(MapShapeLayerDelegate value) {
    if (_mapDelegate == value) {
      return;
    }

    if (_mapDelegate != null &&
        value != null &&
        _mapDelegate.shapeFile != value.shapeFile) {
      _mapDelegate = value;
      return;
    }

    _mapDelegate = value;
    _currentSelectedItem = null;
    _prevSelectedItem = null;
    _refresh();
    markNeedsPaint();
    _state.dataLabelAnimationController.value = 0.0;
    _state.bubbleAnimationController.value = 0.0;
    SchedulerBinding.instance
        .addPostFrameCallback(_forwardDataLabelAndBubbleAnimation);
  }

  Color get color => _color;
  Color _color;
  set color(Color value) {
    if (_color == value) {
      return;
    }
    _color = value;
    markNeedsPaint();
  }

  Color get strokeColor => _strokeColor;
  Color _strokeColor;
  set strokeColor(Color value) {
    if (_strokeColor == value) {
      return;
    }
    _strokeColor = value;
    markNeedsPaint();
  }

  double get strokeWidth => _strokeWidth;
  double _strokeWidth;
  set strokeWidth(double value) {
    if (_strokeWidth == value) {
      return;
    }
    _strokeWidth = value;
    markNeedsPaint();
  }

  MapBubbleSettings get bubbleSettings => _bubbleSettings;
  MapBubbleSettings _bubbleSettings;
  set bubbleSettings(MapBubbleSettings value) {
    if (_bubbleSettings == value) {
      return;
    }
    if (_bubbleSettings.minRadius != value.minRadius ||
        _bubbleSettings.maxRadius != value.maxRadius) {
      _bubbleSettings = value;
      _mapDataSource.forEach((String key, _MapModel mapModel) {
        _updateBubbleRadius(mapModel);
      });
    } else {
      _bubbleSettings = value;
    }
    markNeedsPaint();
  }

  MapLegendSettings get legendSettings => _legendSettings;
  MapLegendSettings _legendSettings;
  set legendSettings(MapLegendSettings value) {
    if (_legendSettings == value) {
      return;
    }
    _legendSettings = value;
    markNeedsPaint();
  }

  MapSelectionSettings get selectionSettings => _selectionSettings;
  MapSelectionSettings _selectionSettings;
  set selectionSettings(MapSelectionSettings value) {
    if (_selectionSettings == value) {
      return;
    }
    _selectionSettings = value;
  }

  bool get enableShapeTooltip => _enableShapeTooltip;
  bool _enableShapeTooltip;
  set enableShapeTooltip(bool value) {
    if (_enableShapeTooltip == value) {
      return;
    }
    _enableShapeTooltip = value;
  }

  bool get enableBubbleTooltip => _enableBubbleTooltip;
  bool _enableBubbleTooltip;
  set enableBubbleTooltip(bool value) {
    if (_enableBubbleTooltip == value) {
      return;
    }
    _enableBubbleTooltip = value;
  }

  bool get enableSelection => _enableSelection;
  bool _enableSelection;
  set enableSelection(bool value) {
    if (_enableSelection == value) {
      return;
    }
    _enableSelection = value;
  }

  SfMapsThemeData get mapsThemeData => _mapsThemeData;
  SfMapsThemeData _mapsThemeData;
  set mapsThemeData(SfMapsThemeData value) {
    if (_mapsThemeData == value) {
      return;
    }
    _mapsThemeData = value;
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;
  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    markNeedsLayout();
  }

  void _refresh() {
    _shapeSizeFactor = 1;
    _shapeLayerTranslationOffset = const Offset(0.0, 0.0);
    if (_currentSize != null &&
        _mapDataSource != null &&
        _mapDataSource != null &&
        _mapDataSource.isNotEmpty) {
      _calculateFactor();
      _calculateTranslationPoint();
      _updateMapDataSourceForVisual();
    }
  }

  void _calculateFactor() {
    final Offset minPoint = _getPixelPositionFromCoordinate(
        _state.shapeBounds.minLongitude,
        _state.shapeBounds.minLatitude,
        _currentSize);
    final Offset maxPoint = _getPixelPositionFromCoordinate(
        _state.shapeBounds.maxLongitude,
        _state.shapeBounds.maxLatitude,
        _currentSize);
    _actualShapeHeight = (maxPoint.dy - minPoint.dy).abs();
    _actualShapeWidth = (maxPoint.dx - minPoint.dx).abs();
    _shapeSizeFactor = min(_currentSize.height / _actualShapeHeight,
        _currentSize.width / _actualShapeWidth);
  }

  void _calculateTranslationPoint() {
    const double scale = 1;
    double dx = 0.0, dy = 0.0;
    double minTranslationX, minTranslationY, maxTranslationX, maxTranslationY;

    final Offset minPoint = _getPixelPositionFromCoordinate(
        _state.shapeBounds.minLongitude,
        _state.shapeBounds.minLatitude,
        _currentSize,
        factor: _shapeSizeFactor);
    final Offset maxPoint = _getPixelPositionFromCoordinate(
        _state.shapeBounds.maxLongitude,
        _state.shapeBounds.maxLatitude,
        _currentSize,
        factor: _shapeSizeFactor);

    maxTranslationY = -maxPoint.dy;
    minTranslationY =
        (_currentSize.height - _actualShapeHeight * scale) / scale - 1;
    maxTranslationX = -minPoint.dx;
    minTranslationX =
        (_currentSize.width - _actualShapeWidth * scale) / scale - 1;

    if (dy > maxTranslationY) {
      dy = maxTranslationY;
    } else if (dy < minTranslationY) {
      dy = minTranslationY;
    }

    if (dx > maxTranslationX) {
      dx = maxTranslationX;
    } else if (dx < minTranslationX) {
      dx = minTranslationX;
    }

    final Offset centerTranslation = Offset(
        _currentSize.width - _actualShapeWidth * _shapeSizeFactor,
        _currentSize.height - _actualShapeHeight * _shapeSizeFactor);
    _shapeLayerTranslationOffset =
        Offset(dx + centerTranslation.dx / 2, dy + centerTranslation.dy / 2);
  }

  void _updateMapDataSourceForVisual() {
    if (_mapDataSource != null) {
      Offset point;
      Path shapePath;
      dynamic coordinate;
      List<Offset> pixelPoints;
      List<dynamic> rawPoints;
      int rawPointsLength, pointsLength;
      _mapDataSource.forEach((String key, _MapModel mapModel) {
        double signedArea = 0.0, centerX = 0.0, centerY = 0.0;
        rawPointsLength = mapModel.rawPoints.length;
        mapModel.pixelPoints = List<List<Offset>>(rawPointsLength);
        shapePath = Path();
        for (int j = 0; j < rawPointsLength; j++) {
          rawPoints = mapModel.rawPoints[j];
          pointsLength = rawPoints.length;
          pixelPoints = mapModel.pixelPoints[j] = List<Offset>(pointsLength);
          for (int k = 0; k < pointsLength; k++) {
            coordinate = rawPoints[k];
            point = pixelPoints[k] = _getPixelPositionFromCoordinate(
                coordinate[0], coordinate[1], _currentSize,
                translationOffset: _shapeLayerTranslationOffset,
                factor: _shapeSizeFactor);
            if (k == 0) {
              shapePath.moveTo(point.dx, point.dy);
            } else {
              shapePath.lineTo(point.dx, point.dy);
              final int l = k - 1;
              if ((_state.widget.showDataLabels || _state.widget.showBubbles) &&
                  l + 1 < pixelPoints.length) {
                // Used mathematical formula to find the center of
                // polygon points.
                final double x0 = pixelPoints[l].dx, y0 = pixelPoints[l].dy;
                final double x1 = pixelPoints[l + 1].dx,
                    y1 = pixelPoints[l + 1].dy;
                signedArea += (x0 * y1) - (y0 * x1);
                centerX += (x0 + x1) * (x0 * y1 - x1 * y0);
                centerY += (y0 + y1) * (x0 * y1 - x1 * y0);
              }
            }
          }
          shapePath.close();
        }

        mapModel.shapePath = shapePath;
        if (_state.widget.showDataLabels || _state.widget.showBubbles) {
          // Used mathematical formula to find the center of polygon points.
          signedArea /= 2;
          centerX = centerX / (6 * signedArea);
          centerY = centerY / (6 * signedArea);
          mapModel.shapePathCenter = Offset(centerX, centerY);
          double minX, maxX;
          double distance,
              minDistance = double.infinity,
              maxDistance = double.negativeInfinity;

          final List<double> minDistances = <double>[double.infinity];
          final List<double> maxDistances = <double>[double.negativeInfinity];
          for (final List<Offset> points in mapModel.pixelPoints) {
            for (final Offset point in points) {
              distance = (centerY - point.dy).abs();
              if (point.dx < centerX) {
                // Collected all points which is less 10 pixels distance from
                // 'center y' to position the labels more smartly.
                if (minX != null && distance < 10) {
                  minDistances.add(point.dx);
                }
                if (distance < minDistance) {
                  minX = point.dx;
                  minDistance = distance;
                }
              } else if (point.dx > centerX) {
                if (maxX != null && distance < 10) {
                  maxDistances.add(point.dx);
                }
                if (distance > maxDistance) {
                  maxX = point.dx;
                  maxDistance = distance;
                }
              }
            }
          }

          mapModel.shapeWidth = max(maxX, maxDistances.reduce(max)) -
              min(minX, minDistances.reduce(min));
        }

        _updateBubbleRadius(mapModel);

        if (mapModel.isSelected && mapModel.dataIndex != null) {
          _currentSelectedItem = mapModel;
          _prevSelectedItem = mapModel;
        }
      });
    }
  }

  void _updateBubbleRadius(_MapModel mapModel) {
    final double bubbleSizeValue = mapModel.bubbleSizeValue;
    if (bubbleSizeValue != null) {
      if (bubbleSizeValue == _state.minBubbleValue) {
        mapModel.bubbleRadius = _bubbleSettings.minRadius;
      } else if (bubbleSizeValue == _state.maxBubbleValue) {
        mapModel.bubbleRadius = _bubbleSettings.maxRadius;
      } else {
        final double percentage = ((bubbleSizeValue - _state.minBubbleValue) /
                (_state.maxBubbleValue - _state.minBubbleValue)) *
            100;
        mapModel.bubbleRadius = bubbleSettings.minRadius +
            (bubbleSettings.maxRadius - bubbleSettings.minRadius) *
                (percentage / 100);
      }
    }

    if (_enableBubbleTooltip && mapModel.bubbleRadius != null) {
      final Path bubblePath = Path()
        ..addOval(Rect.fromCircle(
          center: mapModel.shapePathCenter,
          radius: mapModel.bubbleRadius,
        ));
      mapModel.bubblePath = bubblePath;
    }
  }

  /// Returns fill Color for current shape.
  Color _getFillColor(_MapModel model) {
    if (_legendSettings.enableToggleInteraction &&
        _state.legendController._toggledIndices.contains(model.actualIndex)) {
      if (_legendSettings.toggledShapeOpacity != null) {
        return _legendSettings.toggledShapeColor
            .withOpacity(_legendSettings.toggledShapeOpacity);
      } else {
        return _legendSettings.toggledShapeColor;
      }
    }
    return model.shapeColor ?? _color;
  }

  /// Returns stroke paint for current shape.
  Paint _getStrokePaint(int shapeIndex, Paint strokePaint) {
    strokePaint
      ..color = _strokeColor
      ..strokeWidth = _strokeWidth;
    if (_legendSettings.enableToggleInteraction &&
        _state.legendController._toggledIndices.contains(shapeIndex)) {
      if (_legendSettings.toggledShapeStrokeColor != null) {
        strokePaint.color = _legendSettings.toggledShapeStrokeColor;
      }
      if (_legendSettings.toggledShapeStrokeWidth != null) {
        strokePaint.strokeWidth = _legendSettings.toggledShapeStrokeWidth;
      }
    }
    return strokePaint;
  }

  // Invoking animation for data label and bubble.
  void _forwardDataLabelAndBubbleAnimation(Duration timeStamp) {
    if (_state.mounted) {
      if (_state.widget.showDataLabels) {
        _state.dataLabelAnimationController.forward(from: 0);
      } else if (_state.widget.showBubbles) {
        _state.bubbleAnimationController.forward(from: 0);
      }
    }
  }

  void _onTapUp(TapUpDetails details) {
    if (_mapDataSource == null || _mapDataSource.isEmpty) {
      return;
    }

    double bubbleRadius;
    String tooltipText;
    final bool hasController = _state.widget.controller != null;
    final int length = _mapDataSource.values.length;
    for (int i = 0; i < length; i++) {
      final _MapModel mapModel = _mapDataSource.values.elementAt(i);
      final bool isToggled = _state.legendController._toggledIndices
          .contains(mapModel.actualIndex);
      // Selection will work only when given data count.
      // Selection will not work on toggled shapes.
      if (!isToggled &&
          (_enableSelection || _enableShapeTooltip) &&
          mapModel.shapePath.contains(details.localPosition)) {
        if (_enableSelection && mapModel.dataIndex != null) {
          if (hasController) {
            _state.widget.controller.selectedIndex =
                _state.widget.controller.selectedIndex == mapModel.dataIndex
                    ? -1
                    : mapModel.dataIndex;
          } else {
            _currentSelectedItem = mapModel;
            _updateSelectedItem();
          }
        }

        if (!_mouseIsConnected && _enableShapeTooltip) {
          if (_mapDelegate.shapeTooltipTextMapper != null &&
              mapModel.dataIndex != null) {
            tooltipText =
                _mapDelegate.shapeTooltipTextMapper(mapModel.dataIndex);
          } else {
            tooltipText = mapModel.primaryKey;
          }
        }

        if (!_enableBubbleTooltip) {
          break;
        }
      }

      // Considered _mouseIsConnected to ignore
      // tooltip rendering while tapping in web platform.
      // Tooltip will not work on toggled shapes.
      if (!_mouseIsConnected &&
          !isToggled &&
          _enableBubbleTooltip &&
          mapModel.bubblePath != null &&
          mapModel.bubblePath.contains(details.localPosition)) {
        if (bubbleRadius == null) {
          bubbleRadius = mapModel.bubbleRadius;
          if (_mapDelegate.bubbleTooltipTextMapper != null &&
              mapModel.dataIndex != null) {
            tooltipText =
                _mapDelegate.bubbleTooltipTextMapper(mapModel.dataIndex);
          } else {
            tooltipText = mapModel.primaryKey;
          }
        } else if (mapModel.bubbleRadius < bubbleRadius) {
          if (_mapDelegate.bubbleTooltipTextMapper != null &&
              mapModel.dataIndex != null) {
            tooltipText =
                _mapDelegate.bubbleTooltipTextMapper(mapModel.dataIndex);
          } else {
            tooltipText = mapModel.primaryKey;
          }
        }
      }
    }

    if (!_mouseIsConnected && tooltipText != null) {
      RenderBox child = lastChild;
      while (child != null) {
        final _MapChildParentData childParentData = child.parentData;
        if (child is _RenderMapTooltip) {
          final _RenderMapTooltip tooltipRenderBox = child;
          tooltipRenderBox._tooltipText = tooltipText;
          tooltipRenderBox._touchPosition = details.localPosition;
          tooltipRenderBox._forwardTooltipAnimation();
          break;
        }

        child = childParentData.previousSibling;
      }
    }
  }

  void _onSelectionChanged(Duration time) {
    if (_state.widget.onSelectionChanged != null) {
      _state.widget.onSelectionChanged(_currentSelectedItem.isSelected
          ? _currentSelectedItem.dataIndex
          : -1);
    }
  }

  void _updateSelectedItem() {
    _currentSelectedItem.isSelected = !_currentSelectedItem.isSelected;
    if (_prevSelectedItem != null &&
        _currentSelectedItem.dataIndex != _prevSelectedItem.dataIndex) {
      _prevSelectedItem.isSelected = false;
    }

    _prevSelectedItem = _currentSelectedItem;
    SchedulerBinding.instance.addPostFrameCallback(_onSelectionChanged);
    markNeedsPaint();
  }

  void _shapeLayerControllerListener() {
    if (_enableSelection) {
      final int selectedIndex = _state.widget.controller.selectedIndex;
      assert(selectedIndex < _mapDelegate.dataCount);
      if (selectedIndex == -1) {
        if (_prevSelectedItem != null) {
          _prevSelectedItem.isSelected = false;
        }

        if (_currentSelectedItem != null) {
          _currentSelectedItem.isSelected = false;
        }

        SchedulerBinding.instance.addPostFrameCallback(_onSelectionChanged);
        markNeedsPaint();
        return;
      }

      _currentSelectedItem = _mapDataSource.values.firstWhere(
          (_MapModel element) => element.dataIndex == selectedIndex);
      _updateSelectedItem();
    }
  }

  @override
  void attach(PipelineOwner owner) {
    super.attach(owner);
    _state.widget.controller?.addListener(_shapeLayerControllerListener);
    _state.legendController?.addListener(markNeedsPaint);
    SchedulerBinding.instance
        .addPostFrameCallback(_forwardDataLabelAndBubbleAnimation);
  }

  @override
  void detach() {
    _state.dataLabelAnimationController.value = 0.0;
    _state.bubbleAnimationController.value = 0.0;
    _state.widget.controller?.removeListener(_shapeLayerControllerListener);
    _state.legendController?.removeListener(markNeedsPaint);
    super.detach();
  }

  @override
  bool hitTestSelf(Offset position) => isInteractive;

  @override
  void handleEvent(PointerEvent event, HitTestEntry entry) {
    if (isInteractive && event.down && event is PointerDownEvent) {
      _tapGestureRecognizer.addPointer(event);
    }
  }

  @override
  void setupParentData(RenderObject child) {
    if (child.parentData is! _MapChildParentData) {
      child.parentData = _MapChildParentData();
    }
  }

  @override
  void performLayout() {
    _currentSize = _getActualSize(constraints);
    if (!hasSize || size != _currentSize) {
      size = _currentSize;
      _refresh();
    }

    final BoxConstraints looseContraints = BoxConstraints.loose(_currentSize);
    RenderBox child = firstChild;
    while (child != null) {
      final _MapChildParentData childParentData = child.parentData;
      child.layout(looseContraints, parentUsesSize: true);
      childParentData.mapDataSource = _mapDataSource;
      if (child is _RenderMapMarker) {
        final _RenderMapMarker marker = child;
        childParentData.offset = _getPixelPositionFromCoordinate(
            marker._longitude, marker._latitude, _currentSize,
            translationOffset: _shapeLayerTranslationOffset,
            factor: _shapeSizeFactor);
        if (marker.child != null) {
          childParentData.offset -=
              Offset(child.size.width / 2, child.size.height / 2);
        }
      }

      child = childParentData.nextSibling;
    }
  }

  @override
  void paint(PaintingContext context, Offset offset) {
    if (_mapDataSource != null && _mapDataSource.isNotEmpty) {
      context.canvas.save();
      context.canvas.translate(offset.dx, offset.dy);
      final Paint fillPaint = Paint()..isAntiAlias = true;
      Paint strokePaint = Paint()
        ..isAntiAlias = true
        ..style = PaintingStyle.stroke;
      _mapDataSource.forEach((String key, _MapModel model) {
        fillPaint.color = _getFillColor(model);
        strokePaint = _getStrokePaint(model.actualIndex, strokePaint);
        context.canvas.drawPath(model.shapePath, fillPaint);
        // When stroke width is zero, it will draw a hairline width.
        if (strokePaint.strokeWidth > 0.0) {
          context.canvas.drawPath(model.shapePath, strokePaint);
        }
      });

      // Painting selected shape path.
      if (_currentSelectedItem != null &&
          _currentSelectedItem.isSelected &&
          !_state.legendController._toggledIndices
              .contains(_currentSelectedItem.actualIndex)) {
        fillPaint.color = _mapsThemeData.selectionColor
            .withOpacity(_selectionSettings.opacity);
        strokePaint
          ..color = _mapsThemeData.selectionStrokeColor
          ..strokeWidth = _mapsThemeData.selectionStrokeWidth;
        context.canvas.drawPath(_currentSelectedItem.shapePath, fillPaint);
        if (strokePaint.strokeWidth > 0.0) {
          context.canvas.drawPath(_currentSelectedItem.shapePath, strokePaint);
        }
      }

      context.canvas.restore();

      if (_enableBubbleTooltip || _enableShapeTooltip) {
        RenderBox child = lastChild;
        while (child != null) {
          final _MapChildParentData childParentData = child.parentData;
          if (child is _RenderMapTooltip) {
            childParentData.parentOffset = offset;
            break;
          }

          child = childParentData.previousSibling;
        }
      }

      if (childCount > 0) {
        super.paint(context, offset);
      }
    }
  }
}
